#!/bin/bash

sphinx-apidoc -o source --no-toc ../xGPR ../xGPR/cg_toolkit ../xGPR/scoring_toolkit ../xGPR/constants ../xGPR/random_feature_generation ../xGPR/fitting_toolkit ../xGPR/kernels ../xGPR/optimizers ../xGPR/preconditioners ../xGPR/static_layers ../xGPR/data_handling ../xGPR/tuning_toolkit
